"""Visibility history module."""
# %% Imports
from __future__ import annotations

# Third Party Imports
from intervaltree import IntervalTree
from numpy import ndarray, zeros

# satvis Imports
from satvis.visibility_func import visibilityFunc, zeroCrossingFit


# %% getVisibility
def getVisHist(
    targets: list[dict],
    sensors: list[dict],
    x_targets: ndarray,
    x_sensors: ndarray,
    time: list,
    planet_radius: float,
) -> tuple[IntervalTree, ndarray]:
    """Generate visibility function history between sensors and targets.

    Args:
        targets (`list[dict]`): N-length list of dicts. Each dict must include
            an 'id' field.
        sensors (`list[dict]`): M-length list of dicts. Each dict must include
            an 'id' field.
        x_targets (`ndarray`): [T x 6 x N] State history of targets. The 1st-
            dimension of the array is the [6x1] ECI state vector [position,
            velocity] in km and km/s, respectively.
        x_sensors (`ndarray`): [T x 6 x M] State history of sensors. The 1st-
            dimension of the array is the [6x1] ECI state vector [position,
            velocity] in km and km/s, respectively.
        time (`list`): [T x 1] list of times corresponding to state histories.
        planet_radius (`float`): assume spherical, km

    Returns:
        rise_set_tree (`IntervalTree`): `IntervalTree` instance of class
            representing all intervals for which target-sensor pairs
            can see each other. If no target-sensor pairs can see each
            other during the input time window, the `IntervalTree` is empty.
            The data field of each entry (`rise_set_tree[#].data`) is a dict
            with the following keys:
                {
                    "target_id": target_id,
                    "sensor_id": sensor_id
                }
        vis (`ndarray`): [M x N x T] array of visibility function values for
            all target-sensor pairs for all time.

    Notation:
        N = Number of targets
        M = Number of sensors
        T = Length of time vector
    """
    num_sensors = len(sensors)
    num_targets = len(targets)

    # preallocate visibility array and list of intervals
    vis = zeros([num_sensors, num_targets, len(time)])
    rise_set_ints = []

    # Calculate visibility function values of all sensor-target pairs
    # loop through sensor-platform pairs
    counter = 0
    for i_sensor in range(num_sensors):
        for i_sat in range(num_targets):
            # sensor-target pair name for labelling interval tree
            pair_name = getPairName(targets[i_sat], sensors[i_sensor])

            # loop through time to calc visibility function
            for i_time in range(len(time)):
                r1 = x_sensors[i_time, :3, i_sensor]
                r2 = x_targets[i_time, :3, i_sat]

                # calc visibility function (ignore supplemental outputs)
                [vis[i_sensor, i_sat, i_time], _, _, _] = visibilityFunc(
                    r1=r1,
                    r2=r2,
                    RE=planet_radius,
                    hg=0,
                )

            _, _, new_tree = zeroCrossingFit(
                vis[i_sensor, i_sat],
                time,
                pair_name,
            )

            # extend list of Intervals (note Intervals are not same as
            # IntervalTree)
            rise_set_ints.extend(list(new_tree))

            counter += 1
    rise_set_tree = IntervalTree(rise_set_ints)
    return rise_set_tree, vis


def getPairName(target: dict, sensor: dict) -> dict:
    """Create a target-sensor pair ID dict.

    Args:
        target (`dict`): Target object
        sensor (`dict`): Sensor object

    Returns:
        pair_name (`dict`): {'target_id': target_id, 'sensor_id': sensor_id}
    """
    target_id = target["id"]
    sensor_id = sensor["id"]

    pair_name = {"target_id": target_id, "sensor_id": sensor_id}
    return pair_name



"""Interval Tree Converter module.

Converts list of IntervalTrees into a a format that can be passed to
plotSchedule for schedule availability plots.
"""

# %% Imports
from __future__ import annotations

# Standard Library Imports
from typing import Tuple

# Third Party Imports
from intervaltree import IntervalTree

# %% Function


def intTree2WindowList(schedule_tree: IntervalTree) -> Tuple[list, list, list]:
    """Converts an `IntervalTree`s to an ingestible for `plotSchedule()`.

    Notation:
        M = number of sensors
        N = number of targets
        d = number of availability windows for a given sensor-target pair

    Args:
        schedule_tree (`IntervalTree`): Intervals of sensor-target pair availability.
            The data field in each `Interval` must be a `dict` with the following keys:
                {
                    "target_id": unique_name,
                    "sensor_id": unique_name,
                }

    Returns:
        windows (`list`): [M x N x d] list of lists of lists of pairs
            of tuples, where each nested list is a pair of tuples
            representing the time at the start of an availability window, and
            the duration of that window. d can be any length for each `m` and `n`,
            so long as each entry is a pair of tuples. N must be consistent
            for each M.
        sensor_ids (`list`): List of IDs of sensors that have availability.
        target_ids (`list`): List of IDs of targets that have availability.

    Note that all target and sensor names in `schedule_tree` must be unique.
    """
    # convert IntervalTree to list of Intervals
    list_tree = list(schedule_tree.items())

    # get targets/sensors for each Interval
    target_vec = [interval.data["target_id"] for interval in list_tree]
    sensor_vec = [interval.data["sensor_id"] for interval in list_tree]

    # get unique lists of target/sensor names
    target_ids = list(set(target_vec))
    sensor_ids = list(set(sensor_vec))

    # get number of unique targets/sensors
    num_targets = len(target_ids)
    num_sensors = len(sensor_ids)

    # get start times and durations of each interval
    start_vec = [interval.begin for interval in list_tree]
    dur_vec = [interval.end - interval.begin for interval in list_tree]

    # Assemble inputs for schedule plot
    # initialize availability windows (list of lists)
    windows = [[[] for j in range(num_targets)] for i in range(num_sensors)]

    for i, sens_name in enumerate(sensor_ids):
        # print(f'i={i}, sens_name={sens_name}')
        # get indices of Intervals with with given sensor
        indices = [ctr for ctr, x in enumerate(sensor_vec) if x == sens_name]
        # print(f'indices={indices}')

        # loop through all targets for each sensor
        for j, targ_name in enumerate(target_ids):
            # print(f'j={j}, targ_name={targ_name}')

            # get indices of Intervals with given target
            indices_target = [ctr for ctr, x in enumerate(target_vec) if x == targ_name]

            # get intersection of target-indices and sensor-indices
            intersection = [item for item in indices if item in indices_target]
            # print(f'intersection={intersection}')

            # Next, need to assign intervals that have sense_name in the
            # data field to windows.
            list_of_starts = list(map(start_vec.__getitem__, intersection))
            list_of_durs = list(map(dur_vec.__getitem__, intersection))

            list_of_sets = [[] for x in range(len(list_of_starts))]
            for k, (start, dur) in enumerate(zip(list_of_starts, list_of_durs)):
                list_of_sets[k] = (start, dur)
            windows[i][j] = list_of_sets

    return windows, sensor_ids, target_ids



"""Tests for int_tree_converter module."""
# %% Imports
from __future__ import annotations

# Third Party Imports
import matplotlib.pyplot as plt
from intervaltree import Interval, IntervalTree
from numpy import shape

# satvis Imports
from satvis.int_tree_converter import intTree2WindowList
from satvis.schedule_plots import plotSchedule

# %% Create ScheduleTree()
ivs1 = [(1, 2), (4, 7), (8, 9)]
tree1 = IntervalTree(
    Interval(begin, end, {"target_id": "Sat 1", "sensor_id": "Sens A"})
    for begin, end in ivs1
)
ivs2 = [(4, 6), (8, 11)]
tree2 = IntervalTree(
    Interval(begin, end, {"target_id": "Sat 1", "sensor_id": "Sens B"})
    for begin, end in ivs2
)
tree_merged = tree1 | tree2


# %% Test Converter, simple case
print("\nSimple tests:")

[avail, id_sensors, id_targets] = intTree2WindowList(tree_merged)

print(f"avail = \n{avail}")
print(f"shape(avail) ={shape(avail)}")

# Plot
plt.style.use("default")
fig = plt.figure()
plotSchedule(
    availability=avail,
    target_labels=id_targets,
    sensor_labels=id_sensors,
    fig=fig,
)

# %% Test converter, with single sensor/target
print("\nSingle sensor tests:")
print(f"unavalable sched tree=\n{tree1}")
avail1, id_sensors, id_targets = intTree2WindowList(tree1)
print(f"avail = \n{avail1}")
print(f"shape(avail) ={shape(avail1)}")

# Plot
fig = plt.figure()
plotSchedule(
    availability=avail1,
    target_labels=id_targets,
    sensor_labels=id_sensors,
    fig=fig,
)
# %%
plt.show()
print("done")



"""Test schedule_plots module."""
# %% Imports
from __future__ import annotations

# Third Party Imports
from matplotlib import pyplot as plt

# satvis Imports
from satvis.schedule_plots import padSchedList, plotSchedule

# %% Build data
# building block sets
dat1 = [(2, 2), (6, 3), (10, 5)]
dat2 = [(0, 2), (5, 8)]
dat3 = [(4, 1), (1, 1), (7, 2), (10, 1)]

dat4 = [dat1, dat2, dat3]

dat_a = [(1, 2), (4, 1), (8, 5)]
dat_b = [(0, 4), (4.5, 3)]
dat_c = [(3, 1), (0, 1), (6, 2), (9, 1)]

# multiple sensors, multiple satellites
dat_big = [dat4, [dat_a, dat_b, dat_c]]

dat5 = [[(0, 1)], [(1, 7)], [(8, 6)]]
dat6 = [[(0, 3)], [], [(8, 2)]]
dat_big2 = [dat5, dat6]

dat_empty = [[[], [], []], [[], [], []]]

sensor_labels = ["Sensor A", "Sensor B"]
target_labels = ["Sat 1", "Sat 2", "Sat 3"]

# single sensor, multiple satellite
dat_small = [dat4]
# single satellite, multiple sensors
dat_1sat = [[dat1], [dat2]]
# %% Test pad list
pad = padSchedList(
    dat_1sat,
    [target_labels[0]],
    sensor_labels[:2],
    dat_1sat[1],
    [target_labels[0]],
    [sensor_labels[1]],
)
print(f"superset list=\n{dat_1sat}")
print(f"input list list=\n{dat_1sat[1]}")
print(f"padded list=\n{pad}")

# %% Test Plots
# with no tasked input
print("No tasked inputs")
plt.style.use("default")
f = plt.figure()
f = plotSchedule(dat_big, target_labels, sensor_labels, f)

# with empty tasked input
print("Empty tasked input")
f = plt.figure()
f = plotSchedule(
    dat_big,
    target_labels,
    sensor_labels,
    f,
    dat_empty,
    target_labels,
    sensor_labels,
)

# with populated tasked input
print("populated tasked input")
f = plt.figure()
f = plotSchedule(
    dat_big,
    target_labels,
    sensor_labels,
    f,
    dat_big2,
    target_labels,
    sensor_labels,
)

# with empty availability
print("empty availability")
f = plt.figure()
f = plotSchedule([], target_labels, sensor_labels, f)

# string input for sensor labels
print("string input for sensor labels")
f = plt.figure()
f = plotSchedule(dat_small, target_labels, sensor_labels[0], f)


# one sensor, multiple satellites
print("one sensor, multiple satellites")
f = plt.figure()
f = plotSchedule(dat_small, target_labels, [sensor_labels[0]], f)


# multiple sensors, one satellite
print("multiple sensors, one satellite")
f = plt.figure()
f = plotSchedule(dat_1sat, [target_labels[0]], sensor_labels[:2], f)


# one sensor, one satellite
print("one sensor, one satellite")
f = plt.figure()
f = plotSchedule([dat_1sat[0]], [target_labels[0]], sensor_labels[:1], f)

# one sensor can't see one satellite, other sensor can
avail1 = [[[(2, 1)], [(4, 1)]], [[], [(2, 3)]]]
f = plt.figure()
f = plotSchedule(avail1, ["1", "2"], ["A", "B"], f)


# %% Complicated Inputs
# different num sensors for availability vs tasked
print("different num sensors for  availability vs scheduled")
dat_scheduled = [[[(0, 1), (6, 2)]]]
f = plt.figure()
f = plotSchedule(
    dat_1sat,
    [target_labels[0]],
    sensor_labels[:2],
    f,
    dat_scheduled,
    [target_labels[0]],
    [sensor_labels[1]],
)


# change sensor label
print("different num sensors for  availability vs scheduled (alternate)")
f = plt.figure()
f = plotSchedule(
    dat_1sat,
    [target_labels[0]],
    sensor_labels[:2],
    f,
    dat_scheduled,
    [target_labels[0]],
    [sensor_labels[0]],
)

# %%
print("Schedule has same # of sensors, different satellites")
set_a = [dat1, dat2]
set_tot = [set_a, [dat3, []]]
set_partial = [[dat1], [dat3]]

f = plt.figure()
f = plotSchedule(
    set_tot,
    ["sat1", "sat2"],
    ["sensA", "sensB"],
    f,
    set_partial,
    ["sat1"],
    ["sensA", "sensB"],
)


# %%
plt.show()
print("done")



"""Tests for `visibility_func.py`."""
# %% Imports
from __future__ import annotations

# Third Party Imports
from matplotlib import pyplot as plt
from numpy import array, linspace, sin, zeros

# satvis Imports
from satvis.visibility_func import visibilityFunc, zeroCrossingFit

# %% Test visibilityFunc
print("\n visibilityFunc simple tests...")
RE = 6378
hg = 0
r1 = array([[RE + 400, 0, 0]]).transpose()
r2 = array([[RE, 0, 0]]).transpose()

[v, phi, a1, a2] = visibilityFunc(r1, r2, RE, hg)
print("phi = " + str(phi))
print("type(phi) = " + str(type(phi)))
print("v = " + str(v))
print(f"type(r1) = {type(r1)}")

r1 = array([[1000, 0, 0]]).transpose()
r2 = array([[2000, 0, 0]]).transpose()
[v, phi, a1, a2] = visibilityFunc(r1, r2, 1000, 0)
print(v)
# %% Warning tests
print("\n visibilityFunc warning tests...")
# Check for object being slightly below surface of Earth-- numerically assume
# object is on surface.
r1_alt = array([6378.136299999999, 0, 0])
r2_alt = array([0, 6378.136299999999, 0])
visibilityFunc(r1_alt, r2_alt, 6378.1363, 0)

# Input object below surface of Earth, should display a warning
r1_alt = array([RE - 0.1, 0, 0])
r2_alt = array([0, RE, 0])
[v, phi, a1, a2] = visibilityFunc(r1_alt, r2_alt, RE, 0)
print(f"v = {v}, phi={phi}, alpha1={a1}, alpha2={a2}")

r1_alt = array([RE, 0, 0])
r2_alt = array([0, RE - 0.1, 0])
[v, phi, a1, a2] = visibilityFunc(r1_alt, r2_alt, RE, 0)
print(f"v = {v}, phi={phi}, alpha1={a1}, alpha2={a2}")

# %% Specific bug point tests
print("\n Point tests...")
# these inputs found to have errors previously, so run check to ensure bug was
# fixed.
r1 = array([[8800, 8800, 8800]]).transpose()
r2 = array([[-8000, -8000, -8000]]).transpose()
visibilityFunc(r1, r2, 6378, 0)

r1 = array([[41569.73845258, 6711.91401374, 0]]).transpose()
r2 = array([[41569.73845258, 6711.91401374, 0]]).transpose()

v, phi, alpha1, alpha2 = visibilityFunc(r1, r2, 6371, 0)

print(v)
# %% Test zeroCrossingFit
print("Test zeroCrossingFit...\n")
plt.style.use("default")

# test special cases
t = array([0, 1, 2, 3, 4])
# no visibility window
vis1 = array([-1, -0.9, -0.8, -0.01, -0.1])
[crossings, riseSet, visTree] = zeroCrossingFit(vis1, t, "der")
print("\n No vis windows")
print(f"crossings={crossings} \nriseSet={riseSet}")
print(f"tree={visTree}")

[fig, ax] = plt.subplots()
ax.plot(crossings, zeros(len(crossings)), marker="*", linestyle="None")
ax.plot(crossings, riseSet, marker="|", linestyle="None")
ax.plot(t, vis1)
ax.set_title("No visibility windows")

# visible for whole series
vis2 = -1 * vis1
[crossings, riseSet, visTree] = zeroCrossingFit(vis2, t, "der")
print("\n visibile for whole series")
print(f"crossings={crossings} \nriseSet={riseSet}")
print(f"tree={visTree}")

[fig, ax] = plt.subplots()
ax.plot(crossings, zeros(len(crossings)), marker="*", linestyle="None")
ax.plot(crossings, riseSet, marker="|", linestyle="None")
ax.plot(t, vis2)
ax.set_title("visible for whole series")

# 0-crossing, ends visible
vis3 = array([-1, -0.1, 0.5, 4, 2])
[crossings, riseSet, visTree] = zeroCrossingFit(vis3, t, "der")
print("\n ends visible")
print(f"crossings={crossings} \nriseSet={riseSet}")
print(f"tree={visTree}")

[fig, ax] = plt.subplots()
ax.plot(crossings, zeros(len(crossings)), marker="*", linestyle="None")
ax.plot(crossings, riseSet, marker="|", linestyle="None")
ax.plot(t, vis3)
ax.set_title("w/ 0-crossings, ends visible")


# 0-crossing, ends not visible
vis4 = array([1, 0.1, -0.5, -4, -2])
[crossings, riseSet, visTree] = zeroCrossingFit(vis4, t, "der")
print("\n ends not visible")
print(f"crossings={crossings} \nriseSet={riseSet}")
print(f"tree={visTree}")

[fig, ax] = plt.subplots()
ax.plot(crossings, zeros(len(crossings)), marker="*", linestyle="None")
ax.plot(crossings, riseSet, marker="|", linestyle="None")
ax.plot(t, vis4)
ax.set_title("w/ 0-crossings, ends not-visible")


# starts stradling 0 (-++++)
vis_a = array([-1, 0.2, 0.5, 4, 2])
[crossings, riseSet, visTree] = zeroCrossingFit(vis_a, t, "der")
print("\n straddle, start negative (-++++)")
print(f"crossings={crossings} \nriseSet={riseSet}")
print(f"tree={visTree}")

[fig, ax] = plt.subplots()
ax.plot(crossings, zeros(len(crossings)), marker="*", linestyle="None")
ax.plot(crossings, riseSet, marker="|", linestyle="None")
ax.plot(t, vis_a)
ax.set_title("starts stradling 0 (-++++)")


# starts stradling 0 (+----)
vis_a = array([1, -0.2, -0.5, -4, -2])
[crossings, riseSet, visTree] = zeroCrossingFit(vis_a, t, "der")
print("\n straddle, start positive (+----)")
print(f"crossings={crossings} \nriseSet={riseSet}")
print(f"tree={visTree}")

[fig, ax] = plt.subplots()
ax.plot(crossings, zeros(len(crossings)), marker="*", linestyle="None")
ax.plot(crossings, riseSet, marker="|", linestyle="None")
ax.plot(t, vis_a)
ax.set_title("starts stradling 0 (+----)")


# starts stradling 0 (--+++)
vis_a = array([-1, -0.2, 0.5, 4, 2])
[crossings, riseSet, visTree] = zeroCrossingFit(vis_a, t, "der")
print("\n straddle, start negative (--+++)")
print(f"crossings={crossings} \nriseSet={riseSet}")
print(f"tree={visTree}")

[fig, ax] = plt.subplots()
ax.plot(crossings, zeros(len(crossings)), marker="*", linestyle="None")
ax.plot(crossings, riseSet, marker="|", linestyle="None")
ax.plot(t, vis_a)
ax.set_title("starts stradling 0 (--++++)")


# %% Long time vector
print("\nLong tests \n")
tLong = linspace(0, 30, num=100)
# multiple crossings, ends not visible
vis5 = sin(tLong)
[crossings, riseSet, visTree] = zeroCrossingFit(vis5, tLong, "der")
print("\n Multiple crossings, ends not visible")
print(f"crossings={crossings} \nriseSet={riseSet}")
print(f"tree={visTree}")

[fig, ax] = plt.subplots()
ax.plot(crossings, zeros(len(crossings)), marker="*", linestyle="None")
ax.plot(crossings, riseSet, marker="|", linestyle="None")
ax.plot(tLong, vis5)
ax.set_title("multiple crossings, ends not-visible")


# multiple crossings, ends visible
vis6 = -sin(tLong)
[crossings, riseSet, visTree] = zeroCrossingFit(vis6, tLong, "der")
print("\n Multiple crossings, ends visible")
print(f"crossings={crossings} \nriseSet={riseSet}")
print(f"tree={visTree}")

[fig, ax] = plt.subplots()
ax.plot(crossings, zeros(len(crossings)), marker="*", linestyle="None")
ax.plot(crossings, riseSet, marker="|", linestyle="None")
ax.plot(tLong, vis6)
ax.set_title("multiple crossings, ends visible")

# %%
# Check IntervalTree
vis7 = array([-1, -3, -4, -5, -6])
[crossings3, riseSet3, visTree3] = zeroCrossingFit(
    vis7,
    linspace(0, 5, num=5),
    "hey",
)
print(visTree3)

# %% README Example
print("\n Readme example...")
t = array([0, 1, 2, 3, 4])  # time vector
vis1 = array([-1, -0.1, 0.5, 4, 2])
vis2 = array([-2, -1, -0.5, 1, 1.1])
[_, _, vis_tree1] = zeroCrossingFit(vis1, t, "pair1")
[_, _, vis_tree2] = zeroCrossingFit(vis2, t, "pair2")
combined_tree = vis_tree1 | vis_tree2
print(vis_tree1)
print(vis_tree2)
print(combined_tree)
# %%
plt.show()

print("done")



"""Calculate visibility between two points near a spherical body."""
# %% Imports
from __future__ import annotations

# Standard Library Imports
from typing import Tuple
from warnings import warn

# Third Party Imports
from intervaltree import Interval, IntervalTree
from numpy import (
    append,
    arange,
    arccos,
    array,
    dot,
    isnan,
    isreal,
    nan,
    ndarray,
    sign,
)
from numpy.linalg import norm
from numpy.polynomial import Polynomial


# %% Visibility function
def visibilityFunc(
    r1: ndarray,
    r2: ndarray,
    RE: float,
    hg: float,
) -> Tuple[float, float, float, float]:
    """Calculate visibility function for two position vectors.

    Args:
        r1 (`ndarray`): [3 X 1] ECI position vector of object 1
        r2 (`ndarray`): [3 X 1] ECI position vector of object 2
        RE (`float`): Radius of planet
        hg (`float`): extra height restriction above planet surface

    Returns:
        v (`float`): value of visibility function (positive indicates objects
            can see each other). Returns -1 if either r1 or r1 are below surface
            of planet.
        phi (`float`): angle between position vectors
        alpha1 (`float`): Construction angle 1. Returns numpy.nan if either r1
            or r2 are below surface of planet.
        alpha2 (`float`): Construction angle 2. Returns numpy.nan if either r1
            or r2 are below surface of planet.

    From "Numerical Method for Rapidly Determining Satellite-Satellite
        and Satellite-Ground Station In-View Periods", by Lawton, 1987.
        All argument units are arbitrary distances, just keep consistent.
    """
    # %% Params
    # small value for error threshold
    eps = 1e-13

    # get magnitude of position vectors and radius of body
    RE_prime = RE + hg
    r1_mag = norm(r1)
    r2_mag = norm(r2)

    # %% Correct magnitudes if needed.
    # if points are slightly below Earth surface, change to be on surface.
    if (RE_prime / r1_mag > 1) and (RE_prime / r1_mag < 1 + eps):
        warn("r1_mag nudged to equal RE_prime")
        r1_mag = RE_prime

    if (RE_prime / r2_mag > 1) and (RE_prime / r2_mag < 1 + eps):
        warn("r2_mag nudged to equal RE_prime")
        r2_mag = RE_prime

    # %% Calculate angle between position vectors
    # Check if numerics cause test_var to be >1, and correct if needed
    # Corrects for issues when vectors are close to each other
    test_var = dot(r1.squeeze(), r2.squeeze()) / (r1_mag * r2_mag)
    if test_var > (1 + eps):
        print(f"dot(r1, r2)/(r1_mag * r2_mag)={dot(r1, r2)/(r1_mag * r2_mag)}")
        raise ValueError("dot(r1, r2)/(r1_mag * r2_mag) > 1")
    elif test_var < (-1 - eps):
        print(f"dot(r1, r2)/(r1_mag * r2_mag)={dot(r1, r2)/(r1_mag * r2_mag)}")
        raise ValueError("dot(r1, r2)/(r1_mag * r2_mag) < -1")
    elif test_var > 1 and test_var < (1 + eps):
        test_var = 1
    elif test_var < -1 and test_var > (-1 - eps):
        test_var = -1

    # angle between vectors
    phi = arccos(test_var)

    # %% If either point is far below  surface of body, abort and report not visible
    # check if points are far below surface of body
    r1_flag = False
    if RE_prime / r1_mag > (1 + eps):
        r1_flag = True
        # print(
        #     f"RE_prime/r1_mag={RE_prime/r1_mag}, (RE_prime={RE_prime}, r1_mag={r1_mag})"
        # )
        # raise ValueError("RE_prime > r1_mag")
        warn("RE_prime > r1_mag")

    r2_flag = False
    if RE_prime / r2_mag > (1 + eps):
        r2_flag = True
        # print(
        #     f"RE_prime/r2_mag={RE_prime/r2_mag}, (RE_prime={RE_prime}, r2_mag={r2_mag})"
        # )
        # raise ValueError("RE_prime > r2_mag")
        warn("RE_prime > r2_mag")

    # If either point is far below surface, v and alphas are undefined.
    # However, for v we don't want to return NaN, so just return a negative value
    # let the user know via the warnings above.
    if r1_flag or r2_flag:
        v = -1
        alpha1 = nan
        alpha2 = nan
    else:
        # Nominal path, both points are >= body radius.
        # Calc construction angles.
        alpha1 = arccos(RE_prime / r1_mag)
        alpha2 = arccos(RE_prime / r2_mag)
        v = alpha1 + alpha2 - phi

    # # the checks for r_mag and test_var should ensure that v is a real number,
    # # but just in case, check again
    # if isnan(v):
    #     print(f"alpha1={alpha1}, alpha2={alpha2}, phi={phi}")
    #     print(f"r1={r1}, r2={r2}")
    #     print(f"r1_mag={r1_mag}, r2_mag={r2_mag}")
    #     raise TypeError("Error: v is NaN")

    return v, phi, alpha1, alpha2


# %% Calculate visibility windows
def zeroCrossingFit(
    v: ndarray,
    t: ndarray,
    id: object = None,
) -> Tuple[ndarray[float], ndarray[float], IntervalTree]:
    """Interpolates visibility windows from sparse visibility data.

        Fit curves around zero-crossings of visibility function.
        Fit cubic polynomials to every 4 points, so loop through t[3] - t[end].
        Outputs interval tree object of visibility windows.

        Based on: Alfano, Salvatore, Negron, David Jr., Moore, Jennifer L.,
        “Rapid Determination of Satellite Visibility Periods,” Journal of
        Astronautical Sciences, Vol. 40, No. 2, 1992
    Args:
        v (`ndarray`): [1 x N] array of floats
        t (`ndarray`): [1 x N] array of floats
        id (`any`): (Optional) Identifier for interval tree

    Returns:
        crossings (`ndarray`): [1 x M] array of times at which the
            visibility function crosses 0
        riseSet (`ndarray`): [1 x M] array of +-1 indicating if associated
            value in crossings is a rise point (+1) or set point (-1)
        tree (`IntervalTree`): Interval tree object of time bounds during
            which visibility function is >1

    Note that the ability to detect rise/set times within the first 2 steps
        of the beginning or end of `t` is sketchy due to the 4-point curve
        fit requirement. Workaround is in-place that does a 1st-order polyfit
        when t[0] and t[1] straddle v=0.
    """
    # initialize
    crossings = []
    riseSet = []
    tree = IntervalTree()
    crossIndx = -1

    # Special case: no-zero crossings, V is positive for all time
    if all(sign(v) == 1):
        temp = Interval(t[0], t[-1], id)
        tree.add(temp)
        # print('special case: visible for whole time vector')
        return crossings, riseSet, tree
    # special case: no-zero crossings, V is negative for all time
    elif all(sign(v) == -1):
        # Note: IntervalTree object does not allow null (empty) intervals
        # print('special case: no visibility windows')
        return crossings, riseSet, tree

    # special case: zero-crossing occurs between t[0] and t[1]
    # 1st-order polyfit exception.
    if sign(v[0]) != sign(v[1]):
        crossings = append(crossings, findCrossing(t[:2], v[:2], 1))
        crossIndx += 1
        riseSet = append(riseSet, riseOrSet(v[1]))
        if riseSet[0] == -1:
            temp = Interval(t[0], crossings[0], id)
            tree.add(temp)

    # iterate through time vector for all other cases
    for i in arange(3, len(t)):
        # print('i =' + str(i))
        if i == 3:
            #  starting visibility sign (+1 or -1)
            startV = sign(v[0])

        # fit with 2 mid-points on either side of zero
        if sign(v[i - 1]) != sign(v[i - 2]):
            crossIndx += 1

            # print('i =' + str(i))
            # grab 4 time values
            tSnapshot = array([t[i - 3], t[i - 2], t[i - 1], t[i]])
            # grab 4 v values
            vSnapshot = array([v[i - 3], v[i - 2], v[i - 1], v[i]])

            crossings = append(crossings, findCrossing(tSnapshot, vSnapshot, 3))

            # determine if zero-crossing was a rise or set time and assign
            # val to vector
            riseSet = append(riseSet, riseOrSet(v[i - 1]))

            # Construct interval tree for times when v > 0 (visibility
            # windows)
            # wait for 2 crossings to appear, where 2nd crossing is a
            # set time
            if len(crossings) > 1 and sign(riseSet[crossIndx]) == -1:
                temp = Interval(
                    crossings[crossIndx - 1], crossings[crossIndx], id
                )
                tree.add(temp)
            # create interval if satellite starts in visibility window
            elif len(crossings) == 1 and startV == 1:
                temp = Interval(t[0], crossings[crossIndx], id)
                tree.add(temp)
                # print('special case: first visibility window start
                #        at t0')

        # create interval if simulation ends while visible
        if i == len(t) - 1:
            # at least one crossing has occurred
            if crossIndx >= 0 and sign(v[i]) == 1:
                temp = Interval(crossings[crossIndx], t[i], id)
                tree.add(temp)

    return crossings, riseSet, tree


def findCrossing(
    t: ndarray,
    v: ndarray,
    order: int,
) -> ndarray[float]:
    """Fits a 3rd order polynomial to 4 points and finds root.

    Args:
        t (`ndarray`): [4,] Time values.
        v (`ndarray`): [4,] Visibility function values.
        order (`int`): Order of polyfit.

    Returns:
        `ndarray[float]`: [1, ] Time of zero-crossing.
    """
    # fit a 3rd-order polynomial
    poly = Polynomial.fit(t, v, order)
    # find roots of polynomial
    polyRoots = poly.roots()
    # discard complex roots
    realRoots = polyRoots[isreal(polyRoots)].real

    # get roots that are in domain (range) of the
    # polynomial (t[i-3] and t[i])
    # values of t where 0-crossings occurred
    inRangeRoots = realRoots[
        (realRoots < poly.domain[1]) & (realRoots > poly.domain[0])
    ]

    return inRangeRoots


def riseOrSet(v_i: float) -> int:
    """Switch case for visibility value.

    Args:
        v_i (`float`): Value of visibility function

    Returns:
        `int`: 1, -1, or 0 for rise, set, or anomaly, respectively.
    """
    if sign(v_i) == 1:
        # rise time
        riseSet = 1
    elif sign(v_i) == -1:
        # set time
        riseSet = -1
    else:
        riseSet = 0  # anomaly-->bad!

    return riseSet


def isVis(
    r1: ndarray,
    r2: ndarray,
    RE: float,
    hg: float = 0,
) -> bool:
    """Shortcut wrapper to for boolean visibility.

    Args:
        r1 (`ndarray`): [3 X 1] ECI position vector of object 1
        r2 (`ndarray`): [3 X 1] ECI position vector of object 2
        RE (`float`): Radius of planet
        hg (`float`): extra height restriction above planet surface

    Returns:
        `bool`: True if `r1` and `r2` are visible to each other.
    """
    v, _, _, _ = visibilityFunc(r1, r2, RE, hg)

    # convert to regular bool from numpy.bool
    return bool(v > 0)



"""Test for vis_history.py module."""
# %% Imports
# Third Party Imports
from numpy import array, zeros

# satvis Imports
from satvis.vis_history import getVisHist

# %% Test getVisHist
print("\nTest getVisHist()...")
RE = 6371  # km

# time vector
t1 = [0, 1, 2, 3]
# create dummy target/sensor dicts
target_dicts = [
    {"id": 1},
    {"id": 2},
    {"id": 3},
]
sensor_dicts = [
    {"id": "A"},
    {"id": "B"},
]
# create dummy state history
states_targets = zeros([len(t1), 6, 3])
states_sensors = zeros([len(t1), 6, 2])

# Build state histories for the following:
# Visible to each other:
#   # Target 1 / Sensor A
#   # Target 2 / Sensor B
#   # Target 3 / Sensor B
# Not visible to each other:
#   # Target 1 / Sensor B
#   # Target 2 / Sensor A
#   # Target 3 / Sensor A

# Positions must be greater than Earth radius to get through
# visibilityFunc error check. Velocities aren't used so set to zero.
states_targets[:, 0, 0] = array([8000, 9000, 10000, 11000])
states_targets[:, 0, 1] = -1 * array([8000, 9000, 10000, 11000])
states_targets[:, 0, 2] = -1 * array([8000, 9000, 10000, 11000])

states_sensors[:, 0, 0] = 1.1 * array([8000, 9000, 10000, 11000])
states_sensors[:, 0, 1] = -1.1 * array([8000, 9000, 10000, 11000])

tree, v = getVisHist(
    target_dicts,
    sensor_dicts,
    states_targets,
    states_sensors,
    t1,
    RE,
)
print(f"tree = {tree}")
print(f"visibility = \n{v}")



"""Schedule plots module."""
# %% Imports
from __future__ import annotations

# Standard Library Imports
import warnings

# Third Party Imports
import matplotlib.patches as mpatches
import matplotlib.pyplot as plt
from matplotlib.cm import get_cmap
from matplotlib.figure import Figure
from numpy import empty


# %% Main Function
def plotSchedule(
    availability: list[list[list[tuple[float]]]],
    target_labels: list[str],
    sensor_labels: list[str],
    fig: Figure,
    scheduled: list[list[list[tuple]]] = None,
    scheduled_targ_labels: list[str] = None,
    scheduled_sensor_labels: list[str] = None,
) -> Figure:
    """Generates schedule availability plot for N targets and M sensors.

    Args:
        availability (`list[list[list[tuple[float]]]]`): [M[N[d]]] 3-deep
            nested `list`. The lowest level contains a `tuple` of two `floats`
            representing the start time and duration, respectively, of an availability
            window. N must be the same for every M (all 1st-level nested lists
            must be the same length). d does not need to be the same value for
            every N (2nd-level nested lists don't need to be the same length).
        target_labels (`list[str]`): N-long list of target names.
        sensor_labels (`list[str]`): M-long list of sensor names.
        fig (`Figure`): A matplotlib `Figure`.
        scheduled (`list`): (Optional) Same format as availability. Defaults
            to `None`.
        scheduled_targ_labels (`list[str]`): (Optional) Q-long list of
            strings of target names. Must be a subset of entries in
            `target_labels`. Q >= N.
        scheduled_sensor_labels (`list[str]`): (Optional) P-long list of
            strings of sensor names. Must be a subset of entries in
            `sensor_labels`. P >= M.

    Returns:
        `Figure`: Schedule availability plot as a `broken_barh` plot
            from matplotlib. Colored blocks are the data from 'availability',
            black bars are the data from 'scheduled'.

    Notation:
        M = number of sensors
        N = number of targets
        P = number of scheduled sensors
        Q = number of scheduled targets
        d = the number of access windows for a sensor-target per

    Notes:
       If `scheduled` is not `None`, then `scheduled_targ_labels` and
        `scheduled_sensor_labels` are required.

    Example Usage:
        f = plt.figure()
        avail = [
                 [[(2, 1)], [(4, 1)]],  # access windows for Sensor A
                 [[], [(2, 3)]]  #  access windows for Sensor B
                ]
        target_labels = ['1', '2']
        sensor_labels = ['A', 'B']
        f = plotSchedule(avail, target_labels, sensor_labels, f)
        plt.show()

    """
    # Warnings
    if availability == []:
        warnings.warn("No schedule availability. Plot not generated. ")
        return

    if type(target_labels) != list:
        warnings.warn("target_labels must be a list")
        return

    if type(sensor_labels) != list:
        warnings.warn("sensor_labels must be a list")
        return

    # Generate schedule availability (colored) bars
    fig = genPlot(availability, target_labels, sensor_labels, fig, "wide", None, None)

    # Generate scheduled (black) bars
    if scheduled is not None:
        # pad scheduled list with empty lists to be same dimensions as availability list
        padded_sched = padSchedList(
            availability,
            target_labels,
            sensor_labels,
            scheduled,
            scheduled_targ_labels,
            scheduled_sensor_labels,
        )

        fig = genPlot(
            padded_sched,
            target_labels,
            sensor_labels,
            fig,
            "thin",
            scheduled_targ_labels,
            scheduled_sensor_labels,
        )

    return fig


# %% Supporting Functions
def padSchedList(
    super_list: list[list[list[tuple]]],
    super_targs: list[str],
    super_sens: list[str],
    small_list: list[list[list[tuple]]],
    small_targs: list[str],
    small_sens: list[str],
) -> list[list[list[tuple]]]:
    """Pads `small_list` to be same dimensions as `super_list`.

    Args:
        super_list (`list[list[list[tuple]]]`): Availability.
        super_targs (`list[str]`): Names of all targets in `super_list`.
        super_sens (`list[str]`): Names of all sensors in `super_list`.
        small_list (`list[list[list[tuple]]]`): Scheduled.
        small_targs (`list[str]`): Names of all targets in `small_list`.
        small_sens (`list[str]`): Names of all sensors in `small_list`.

    Returns:
        `list[list[list[tuple]]]`: Same dimensions as `super_list`, but with
            contents of `small_list` corresponding to specified targets and
            sensors. Entries corresponding to targets/sensors not included
            in `small_list` are empty lists.
    """
    padded_list = [[] for x in range(len(super_list))]
    index_small = 0
    for i, (sens, dat) in enumerate(zip(super_sens, super_list)):
        # print(i, sens, dat)
        if sens in small_sens:
            padded_list[i] = small_list[index_small]
            index_small += 1
        # print(padded_list)

    return padded_list


def genPlot(
    dat: list[list[list[tuple[float]]]],
    target_labels: list[str],
    sensor_labels: list[str],
    fig: Figure,
    flag: str,
    sched_targ_labels: list[str],
    sched_sensor_labels: list[str],
):
    """Workhorse function for plotSchedule(). Generates broken_barh plots.

    Args:
        dat (`list[list[list[tuple[float]]]]`): [M[N[d]]] 3-deep
            nested `list`. The lowest level contains a `tuple` of two `floats`
            representing the start time and duration, respectively, of an availability
            window. N must be the same for every M (all 1st-level nested lists
            must be the same length). d does not need to be the same value for
            every N (2nd-level nested lists don't need to be the same length).
        target_labels (`list[str]`): N-long list of target names.
        sensor_labels (`list[str]`): M-long list of sensor names.
        fig (`Figure`): matplotlib `Figure`.
        flag (`str`): 'wide' or 'thin' to set whether to create wide colored
            bars or thin black bars.
        sched_targ_labels (`list[str]`): Q-long list of
            strings of target names. Must be a subset of entries in
            `target_labels`. Q must be less than N.
        sched_sensor_labels (`list[str]`): P-long list of
            strings of sensor names. Must be a subset of entries in
            `sensor_labels`. P must be less than M.

    Returns:
        `Figure`: Matplolib broken_hbar plot of access windows.
    """
    # pick any colormap
    cm = get_cmap(name="gist_rainbow")

    # number of sets (number of sensors)
    num_sensors = len(sensor_labels)

    # number boxes per set (number of targets)
    num_targets = len(target_labels)

    # bar widths
    w_big = 1
    w_small = w_big / num_targets
    w_vsmall = w_small / 2

    # bar y-offsets
    y_big = 1.5 * w_big
    y_small = w_big / num_targets
    y_vec = empty([num_sensors])

    # set values depending on wide/thin flag
    if flag == "wide":
        # add subfigure to empty figure
        ax = fig.add_subplot(1, 1, 1)

        # extra offset multiplier
        x = 0

        # alpha value (transparency for plot)
        al = 0.5

        # width of a set of bars associated with one sensor
        w_set = w_big / num_targets

        def chooseCol(i):
            return cm(i / num_targets)

    elif flag == "thin":
        # black bars must be overlaid on colored bars (no stand-alone black
        #   bars)
        ax = plt.gca()
        x = 1
        al = 1
        w_set = 0.5 * (w_big / num_targets)

        def chooseCol(i):
            return "black"

    # loop through y-axis (sensors)
    for j, sens in enumerate(sensor_labels):
        # set y-offset per grouping of bars
        y_offset = y_big * j + x * (w_vsmall / 2)

        # save y-offset value for plotting
        y_vec[j] = y_offset

        # loop through each color bar
        for i, targ in enumerate(target_labels):
            if flag == "wide":
                plt.broken_barh(
                    dat[j][i],
                    (y_offset + (i * y_small), w_set),
                    facecolor=chooseCol(i),
                    edgecolor="black",
                    alpha=al,
                )
            elif (
                (flag == "thin")
                and (sens in sched_sensor_labels)
                and (targ in sched_targ_labels)
            ):
                plt.broken_barh(
                    dat[j][i],
                    (y_offset + (i * y_small), w_set),
                    facecolor=chooseCol(i),
                    edgecolor="black",
                    alpha=al,
                )

    ax.set_yticks(y_vec + w_big / 2)
    ax.set_yticklabels(sensor_labels)
    ax.set_xlabel("Time (s)")
    ax.set_ylabel("Sensor ID")
    ax.set_title("Schedule Availability")
    leg1 = plt.legend(target_labels, title="Target ID", loc=1)
    ax.add_artist(leg1)
    if flag == "thin":
        black_patch = mpatches.Patch(color="black", label="Scheduled")
        leg2 = plt.legend(handles=[black_patch], loc=4)
        ax.add_artist(leg2)

    return fig
